"""
This directory holds means to get generic information for specific languages.
"""

# Start ignoring PyUnusedCodeBear
from .Language import Language
from .Language import Languages

from .definitions.Unknown import Unknown
from .definitions.C import C
from .definitions.CPP import CPP
from .definitions.CSharp import CSharp
from .definitions.CSS import CSS
from .definitions.Fortran import Fortran
from .definitions.Golang import Golang
from .definitions.html import HTML
from .definitions.Java import Java
from .definitions.JavaScript import JavaScript
from .definitions.JSP import JSP
from .definitions.Matlab import Matlab
from .definitions.ObjectiveC import ObjectiveC
from .definitions.PHP import PHP
from .definitions.PLSQL import PLSQL
from .definitions.Python import Python
from .definitions.Ruby import Ruby
from .definitions.Scala import Scala
from .definitions.Swift import Swift
from .definitions.Vala import Vala
# Stop ignoring PyUnusedCodeBear
